/*
* File TreeLikelihood.java
*
* Copyright (C) 2010 Remco Bouckaert remco@cs.auckland.ac.nz
*
* This file is part of BEAST2.
* See the NOTICE file distributed with this work for additional
* information regarding copyright ownership and licensing.
*
* BEAST is free software; you can redistribute it and/or modify
* it under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
*  BEAST is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with BEAST; if not, write to the
* Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
* Boston, MA  02110-1301  USA
*/

/* GRJ, Apr/May 2020: I edited code from the BADTRIP package (was TreeLikelihoodPoMoTransmission)
 */

package beast.evolution.likelihood;

import java.lang.reflect.Constructor;
import java.util.*;

import beast.core.Description;
import beast.core.Distribution;
import beast.core.Input;
import beast.core.State;
import beast.core.Input.Validate;
import beast.core.parameter.RealParameter;
import beast.core.util.Log;

import beast.evolution.alignment.AlignmentPoMo;
import beast.evolution.datatype.DataTypePoMo;
import beast.evolution.datatype.Sample;
import beast.evolution.sitemodel.SiteModelInterfacePoMo.Base;
import beast.evolution.sitemodel.SiteModelInterface;
import beast.evolution.sitemodel.SiteModelPoMo;
import beast.evolution.substitutionmodel.DefaultEigenSystem;
import beast.evolution.substitutionmodel.EigenDecomposition;
import beast.evolution.substitutionmodel.EigenSystem;
import beast.evolution.substitutionmodel.PoMoGeneral;
import beast.evolution.tree.NodePoMo;
import beast.evolution.tree.TreePoMo;

import beast.util.Randomizer;

import static beast.evolution.pomocov.DebugTune.errOutMatrix;
import static beast.evolution.pomocov.TransmissionDistributions.medianDelayTime;
import static beast.evolution.pomocov.TransmissionDistributions.medianWithinHostTime;


@Description("Calculates the log-likelihood of deep sequencing data on a PoMo transmission tree given a site model and PoMo substitution model.")
public class PoMoCoVTreeLikelihood extends Distribution {
    
    final public Input<AlignmentPoMo> dataInput = new Input<>("data",
            "sequence PoMo data",
            Validate.REQUIRED);

    final public Input<TreePoMo> treePoMoInput = new Input<>("treePoMo",
            "phylogenetic transmission PoMo tree with sequence data in the leafs and internal nodes",
            Validate.REQUIRED);

    final public Input<SiteModelInterface> siteModelInput = new Input<>("siteModel",
            "site model for leafs in the beast.tree",
            Validate.REQUIRED);

    // GRJ TODO I doubt that it is useful to use anything except DefaultEigenSystem
    public Input<String> eigenSystemClass = new Input<String>("eigenSystem",
            "Name of the class used for creating an EigenSystem",
            DefaultEigenSystem.class.getName());

    public Input<Boolean> useAmbiguitiesInput = new Input<Boolean>("useAmbiguities",
            "flag to indicate leafs that sites containing ambigue states should be handled instead of ignored (the default)",
            false);
    
    public Input<RealParameter> seqErrorInput = new Input<RealParameter>("seqError",
            " starting sequencing error rate, 0 by default",
            Validate.REQUIRED);
    
	public Input<RealParameter> bottleneckInput = new Input<RealParameter>("bottleneck",
            "parameter describing how many Ne generiations the bottleneck corresponds to in terms of drift",
            Validate.REQUIRED);
    
    enum Scaling {none, always, _default};
    public Input<Scaling> scaling = new Input<PoMoCoVTreeLikelihood.Scaling>("scaling",
            "type of scaling to use, one of " + Arrays.toString(Scaling.values()) + ". If not specified, the -beagle_scaling flag is used.",
            Scaling._default, Scaling.values());


	private boolean debugFlag = Boolean.valueOf(System.getProperty("badtrip.debug"));
	// for debugging use           –enableassertions -Dbadtrip.debug="true"
    private int numberofdebugchecks = 0;


    private AlignmentPoMo alignment;
    private PoMoGeneral substitutionModel;
    private SiteModelPoMo.Base siteModel;
    // flag to indicate ascertainment correction should be applied
    private boolean useAscertainedSitePatterns = false;
    // dealing with proportion of site being invariant
    private double proportionInvariant = 0;
    private List<Integer> constantPattern = null;

    private int nTaxa;
    private int nPatterns;
    private int virtualPopSize;
    private int nStates;

    // Parameters. (Others are in tree and site model.)
    private double sequencingError, sequencingErrorStored;
    private double bottleneck, bottleneckStored;

    // calculation engine
    private BeerLikelihoodCorePoMo likelihoodCore;
    private BeagleTreeLikelihoodPoMo beagle;

    // GRJ TODO Remove when confident of Drift2nucs
    private EigenDecomposition eigenDecompositionBottleneck;
    private EigenSystem eigenSystemBottleneck;
    private double[][] rateMatrixBottleneck;
    //transition probabilities for bottleneck: just one set for the whole tree
    private double[] probabilitiesBottleneck, probabilitiesBottleneckStored;

	private EigenDecomposition eigenDecompositionDrift2nucs;
	private EigenSystem eigenSystemDrift2nucs;
	private double[][] rateMatrixDrift2nucs;
    // transition probabilities for part of bottleneck, for any two nucelotides. Used to
    // achieve exactly the same as probabilitiesBottleneck, but more efficiently.
    private double[] probabilitiesDrift2nucs, probabilitiesDrift2nucsStored;

    // For samples. These change when sequencingError changes
    private double[][] partialsTaxa, partialsTaxaStored;

    private static class EventInHost {
        final int chIdx;
        final int smpIdx;
        final double height;
        double tipwardInterval;

        public EventInHost(int chIdx, int smpIdx, double height) {
            assert height >= 0.0;
            this.chIdx = chIdx;
            this.smpIdx = smpIdx;
            this.height = height;
            this.tipwardInterval = 0.0;
        }
    }




    @Override
    public void initAndValidate() throws IllegalArgumentException {
    	
		//NICOLA TODO : try to see if you can use beagle.
        beagle = null;

        //int nodeCount = treePoMoInput.get().getNodeCount();
        if (!((siteModelInput.get()) instanceof SiteModelPoMo.Base)) {
        	throw new IllegalArgumentException("siteModel input should be of type SiteModel.Base");
        }

        alignment = (AlignmentPoMo) dataInput.get();
        siteModel = (Base) siteModelInput.get();
        siteModel.setDataType((DataTypePoMo)(dataInput.get().getDataType()));
        substitutionModel = siteModel.substModelInput.get();
        nTaxa = alignment.getTaxonCount();
        nPatterns = dataInput.get().getPatternCount();
        virtualPopSize = substitutionModel.getPopSize();
        nStates = substitutionModel.getStateCount();
        assert nStates == 4 + 6 * (virtualPopSize - 1);

        proportionInvariant = siteModel.getProportionInvariant();
        siteModel.setPropInvariantIsCategory(false);
        if (proportionInvariant > 0) {
            calcConstantPatternIndices(nPatterns, nStates);
        }

        sequencingError = seqErrorInput.get().getValue();
        sequencingErrorStored = sequencingError;
        partialsTaxa = new double[nTaxa][nPatterns * nStates];
        partialsTaxaStored = new double[nTaxa][nPatterns * nStates];

        if (dataInput.get().isAscertained) {
            useAscertainedSitePatterns = true;
        }

        likelihoodCore = new BeerLikelihoodCorePoMo(nStates);

        Log.info.println(getClass().getSimpleName() + "(" + getID() + ") uses " + likelihoodCore.getClass().getSimpleName());
        Log.info.println("  " + alignment.toString(true));        bottleneck = bottleneckInput.get().getValue();

        bottleneckStored = bottleneck;

        rateMatrixBottleneck = new double[nStates][nStates];
	    rateMatrixBottleneck = getRatesBottleneck(virtualPopSize);
		try {
			eigenSystemBottleneck = createEigenSystem();
		} catch (Exception e) {
			e.printStackTrace();
		}
		eigenDecompositionBottleneck = eigenSystemBottleneck.decomposeMatrix(rateMatrixBottleneck);

        rateMatrixDrift2nucs = new double[virtualPopSize + 1][virtualPopSize + 1];
        rateMatrixDrift2nucs = getDrift2nucsRates(virtualPopSize);
		try {
			eigenSystemDrift2nucs = createDrift2nucsEigenSystem();
		} catch (Exception e) {
			e.printStackTrace();
		}
		eigenDecompositionDrift2nucs = eigenSystemDrift2nucs.decomposeMatrix(rateMatrixDrift2nucs);

        probabilitiesBottleneck = new double[(nStates) * (nStates)];
        Arrays.fill(probabilitiesBottleneck, 1.0);
        getTransitionProbabilitiesBottleneck();
        probabilitiesBottleneckStored = probabilitiesBottleneck.clone();

		probabilitiesDrift2nucs = new double[(virtualPopSize + 1) *(virtualPopSize + 1)];
		Arrays.fill(probabilitiesDrift2nucs, 1.0);
		getTransitionProbabilitiesDrift2nucs();
		probabilitiesDrift2nucsStored = probabilitiesDrift2nucs.clone();

        if (debugFlag) {
            errOutMatrix("probabilitiesBottleneck", probabilitiesBottleneck);
            errOutMatrix("probabilitiesDrift2nucs", probabilitiesDrift2nucs);
        }

        List<String> tNames = alignment.getTaxaNames();
        for (int t = 0; t < nTaxa; t++){
            double[] currPartials = calcSamplePartials(alignment.getTaxonIndex(tNames.get(t)), nPatterns);
        	System.arraycopy(currPartials, 0, partialsTaxa[t], 0, currPartials.length);
    		System.arraycopy(currPartials, 0, partialsTaxaStored[t], 0, currPartials.length);
        }

        initCore();
    }



    /**
     * Calculate the log likelihood of the current state.
     *
     * @return the log likelihood.
     */
    @Override
    public double calculateLogP() throws IllegalArgumentException {
        // Step 1. calculate root partials
        final TreePoMo tree = treePoMoInput.get();
        // Update bottleneck matrix if needed
        bottleneck = bottleneckInput.get().getValue();
        if (bottleneck != bottleneckStored){
            getTransitionProbabilitiesBottleneck();
            getTransitionProbabilitiesDrift2nucs();
        }
        // update sample partials if needed
        sequencingError = seqErrorInput.get().getValue();
        if (sequencingError != sequencingErrorStored){
            List<String> tNames = alignment.getTaxaNames();
            for (int t=0; t<nTaxa; t++){
                double[] currPartials = calcSamplePartials(alignment.getTaxonIndex(tNames.get(t)), nPatterns);
                System.arraycopy(currPartials, 0, partialsTaxa[t], 0, nStates *nPatterns);
            }
        }
        // recursively calculate root partials
        double[] rootPartials = calcPartialsSubtree(tree.getRoot(), nPatterns, tree.getUnobservedCases());

        // Step 2. root partials to patternLLhoods[]
        double[] frequencies = substitutionModel.getFrequencies();
        double[] proportions = siteModel.getCategoryProportions();
        assert proportions.length == 1;
        assert proportions[0] == 1.0;
        // weightedRootPartials[] are weighted per rate category, but this is not yet implemented.
        double[] weightedRootPartials = new double[nPatterns * nStates];
        likelihoodCore.calculateIntegratePartials(rootPartials, proportions, weightedRootPartials);
        if (constantPattern != null) {
            proportionInvariant = siteModel.getProportionInvariant();
            for (final int i : constantPattern) {
                weightedRootPartials[i] += proportionInvariant;
            }
        }
        double [] patternLLhoods = new double[nPatterns];
        likelihoodCore.calculateLogLikelihoods(weightedRootPartials, frequencies, patternLLhoods);

        // Step 3. Weighted sum of patternLLhoods[].
        logP = sumLogPOverPatterns(patternLLhoods);
        return logP;
    }




    /** CalculationNode methods **/

    /**
     * check state for changed variables and update temp results if necessary *
     */
    @Override
    protected boolean requiresRecalculation() {
        return true;
        //System.out.println("Start requiresRecalculation ");

        //NICOLA TODO : allow beagle
    }

    @Override
    public void store() {
        //System.out.println("Start store ");
        if (sequencingError != sequencingErrorStored){
            //double[] currPartials;//= new double[nStateCount];
            //List<String> tNames = alignment.getTaxaNames();
            for (int t=0; t<nTaxa; t++){
                System.arraycopy(partialsTaxa[t], 0, partialsTaxaStored[t], 0, nPatterns* nStates);
            }
        }
        sequencingErrorStored = sequencingError;
        if (beagle != null) {
            beagle.store();
            super.store();
            return;
        }
        if (likelihoodCore != null) {
            likelihoodCore.store();
        }
        super.store();

        /*TODO-Drift2nucs
        if (bottleneck!=bottleneckStored){
        	probabilitiesBottleneckStored=probabilitiesBottleneck.clone();
            probabilitiesDrift2nucsStored = probabilitiesDrift2nucs.clone();
        	bottleneckStored=bottleneck;
        }*/

    }

    @Override
    public void restore() {
        //System.out.println("Start restore ");
        if (sequencingError != sequencingErrorStored){
            for (int t=0; t<nTaxa; t++){
                System.arraycopy(partialsTaxaStored[t], 0, partialsTaxa[t], 0, nPatterns* nStates);
            }
        }
        sequencingError = sequencingErrorStored;
        //partialsDirty=false;
        if (beagle != null) {
            beagle.restore();
            super.restore();
            return;
        }
        if (likelihoodCore != null) {
            likelihoodCore.restore();
        }
        super.restore();

        /*TODO-Drift2nucs
        if (bottleneck!=bottleneckStored){
        	probabilitiesBottleneck = probabilitiesBottleneckStored.clone();
        	probabilitiesDrift2nucs = probabilitiesDrift2nucsStored.clone();
        	bottleneckStored=bottleneck;
        }*/
    }

    /**
     * @return a list of unique ids for the state nodes that form the argument
     */
    public List<String> getArguments() {
        return Collections.singletonList(dataInput.get().getID());
    }

    /**
     * @return a list of unique ids for the state nodes that make up the conditions
     */
    public List<String> getConditions() {
        return siteModel.getConditions();
    }


    /**
     * This method samples the sequences based on the tree and site model.
     */
    @Override
    public void sample(State state, Random random) {
        throw new UnsupportedOperationException("Can't sample a fixed alignment!");
    }




    // *****************************************************************************
    // ***********************      private     ************************************
    // *****************************************************************************

    // *****************************************************************************
    // *****************        Initialisation      ********************************
    // *****************************************************************************



    /**
     * create an EigenSystem of the class indicated by the eigenSystemClass input *
     */
    private EigenSystem createEigenSystem() throws Exception {
        Constructor<?>[] ctors = Class.forName(eigenSystemClass.get()).getDeclaredConstructors();
        Constructor<?> ctor = null;
        for (int i = 0; i < ctors.length; i++) {
            ctor = ctors[i];
            if (ctor.getGenericParameterTypes().length == 1)
                break;
        }
        ctor.setAccessible(true);
        return (EigenSystem) ctor.newInstance(nStates);
    }


    private EigenSystem createDrift2nucsEigenSystem() throws Exception {
        Constructor<?>[] ctors = Class.forName(eigenSystemClass.get()).getDeclaredConstructors();
        Constructor<?> ctor = null;
        for (int i = 0; i < ctors.length; i++) {
            ctor = ctors[i];
            if (ctor.getGenericParameterTypes().length == 1)
                break;
        }
        ctor.setAccessible(true);
        return (EigenSystem) ctor.newInstance(virtualPopSize + 1);
    }


    /**
     * Determine indices of weightedRootPartials[] that need to be updates
     * // due to sites being invariant. If none of the sites are invariant,
     * // the 'site invariant' category does not contribute anything to the
     * // root probability. If the site IS invariant for a certain character,
     * // taking ambiguities in account, there is a contribution of 1 from
     * // the 'site invariant' category.
     */
    private void calcConstantPatternIndices(final int nPatterns, final int nStateCount) {
        constantPattern = new ArrayList<Integer>();
        for (int i = 0; i < nPatterns; i++) {
           // final int[][] pattern = dataInput.get().getPatternPoMo(i);
            final boolean[] bIsInvariant = new boolean[nStateCount];
            Arrays.fill(bIsInvariant, true);
            for (int k = 0; k < nStateCount; k++) {
                if (bIsInvariant[k]) {
                    constantPattern.add(i * nStateCount + k);
                }
            }
        }
    }

    private void initCore() {
    	TreePoMo treePoMo=treePoMoInput.get();
    	//TODO: move to later (calculateLogP) to account for new nodes added to the tree
        final int nodeCount = treePoMo.getNodeCount();
        likelihoodCore.initialize(
                nodeCount,
                nPatterns,
                siteModel.getCategoryCount(),
                true, useAmbiguitiesInput.get()
        );
    }


    // *****************************************************************************
    // ******************** Main log lhood calculation *****************************
    // *****************************************************************************

    
    /**
     * Recursively calculate partials below node*
     */
    private double[] calcPartialsSubtree(NodePoMo node, int patternCount, boolean unobservedCases) {
        ArrayList<double[]> childrensPartials = new ArrayList<double[]>();
        double[] partial;
        for (NodePoMo c : node.getChildren()){
            partial = calcPartialsSubtree(c, patternCount, unobservedCases);
            childrensPartials.add(partial);
        }

        numberofdebugchecks++;

        double[] partials = calcNodePartials(node, childrensPartials, patternCount, unobservedCases);
        return partials;
    }


    private void checkPartials(double[] partial, String message) {
        if (debugFlag) {
            double sum=0.0;
            for (int i=0; i<partial.length;i++) sum+=partial[i];
            if (sum == 0.0) {
                System.out.println(message);
            }
            assert sum > 0.0;
        }
	}


    private ArrayList<EventInHost> collectEventsInHost(NodePoMo node, int nSamples, int nChildren) {
        ArrayList<EventInHost> EIHs = new ArrayList<>();
        int nIntervals = nChildren + nSamples;
        assert nIntervals > 0;
        for (int ch = 0; ch < nChildren; ch++) {
            EventInHost EIH = new EventInHost(ch, -1, node.getChild(ch).getHeight());
            EIHs.add(EIH);
        }
        for (int smp = 0; smp < nSamples; smp++) {
            EventInHost EIH = new EventInHost(-1, smp, node.getSample(smp).time);
            EIHs.add(EIH);
        }
        EventInHost EIH = new EventInHost(-1, -1, node.getHeight());
        EIHs.add(EIH);
        Collections.sort(EIHs, Comparator.comparing(item -> item.height));
        for (int eih = 1; eih < EIHs.size(); eih++) {
            EIHs.get(eih).tipwardInterval = EIHs.get(eih).height - EIHs.get(eih-1).height;
        }
        return  EIHs;
    }


    private double[] calcNodePartials(NodePoMo node, ArrayList<double[]> childrensPartials, int patternCount, boolean unobservedCases) {
        List<Sample> samples = node.getSamples();
    	int nSamples = samples.size();
        int nChildren = node.getChildren().size();
        int nIntervals = nChildren+nSamples;
        if (nIntervals == 0){
        	double[] partial1 = new double[likelihoodCore.partialsSize];
        	Arrays.fill(partial1, 1.0);
        	return partial1;
        }
        ArrayList<EventInHost> EIHs = collectEventsInHost(node, nSamples, nChildren);

        double[] transProbsMatrix = new double[likelihoodCore.matrixSize];
        double[] partials = new double[likelihoodCore.partialsSize];
     	double[] newPartials = new double[likelihoodCore.partialsSize];
        EventInHost EIH = EIHs.get(0);
        assert EIH.tipwardInterval == 0.0;
        // Start with the most tipward child or sample, get partials[]
        if (EIH.chIdx >= 0) {
            System.arraycopy(childrensPartials.get(EIH.chIdx), 0, partials, 0, patternCount* nStates);
            checkPartials(partials, "1. most recent event (child)");
        } else {
            assert EIH.smpIdx >= 0;
            int taxonIdx = alignment.getTaxonIndex(samples.get(EIH.smpIdx).getTaxon().getID());
            System.arraycopy(partialsTaxa[taxonIdx], 0, partials, 0, patternCount* nStates);
            checkPartials(partials, "2. most recent event (sample)");
        }
        // Intervals between children and samples. For each interval
        // Have p = partials[] from previous step
        // Make M = transProbsMatrix[,].
        // Get q = chOrSmpPartials[] = child partials or sample partials
        // newPartials[] = Mp * q where Mp is matrix-vector, and * is elementwise
        // copy newPartials[] to partials[]
        final double branchRate = siteModel.getRateForCategory(0, null);
        for (int eih = 1; eih < EIHs.size() - 1; eih++) {
            EIH = EIHs.get(eih);
            substitutionModel.getTransitionProbabilities(EIH.tipwardInterval, branchRate, transProbsMatrix);
            double [] chOrSmpPartials;
            if (EIH.chIdx >= 0) {
                chOrSmpPartials = childrensPartials.get(EIH.chIdx);
            } else {
                assert EIH.smpIdx >= 0;
                int taxonIdx = alignment.getTaxonIndex(samples.get(EIH.smpIdx).getTaxon().getID());
                chOrSmpPartials =partialsTaxa[taxonIdx];
            }
            likelihoodCore.calculatePartialsFixedPruning(partials, transProbsMatrix, chOrSmpPartials, newPartials);
            System.arraycopy( newPartials, 0, partials, 0, newPartials.length );
            checkPartials(newPartials, "3. a " + ((EIH.chIdx >= 0) ? "child" : "sample"));
        }
        substitutionModel.getTransitionProbabilities(EIHs.get(EIHs.size() - 1).tipwardInterval, branchRate, transProbsMatrix);
        likelihoodCore.calculatePartialsPruning(partials, transProbsMatrix, newPartials);
        System.arraycopy( newPartials, 0, partials, 0, newPartials.length );
        checkPartials(newPartials, "4. last interval to node");

        if (unobservedCases  &&  !node.isRoot()) {
            double duration = node.getOriginHeight() - node.getHeight();
            calcUnobservenChainPartials(duration, node.getUnobservedCount(), branchRate, partials, transProbsMatrix, newPartials);
        } else {
            calcBottleneckPartials(partials, newPartials);
        }

        return newPartials;
    }



    private void calcBottleneckPartials(double [] partials, double [] newPartials) {
        likelihoodCore.calculatePartialsPruning(partials, probabilitiesBottleneck, newPartials);
        checkPartials(newPartials, "5. bottleneck");
        // GRJ TODO Should be a speedup. 2020-05-08 tested, still cautious!
        double[] D2nPartials = new double[likelihoodCore.partialsSize];
        likelihoodCore.calculatePartialsDrift2NucsBottleneck(partials, probabilitiesDrift2nucs, D2nPartials, substitutionModel.getPopSize());
        checkPartialsMatch(D2nPartials, newPartials);
    }


    private void calcUnobservenChainPartials(double duration, int uc, double branchRate,
                                             double [] partials, double [] transProbsMatrix,  double [] newPartials) {
        double dM = medianDelayTime();
        double wM = medianWithinHostTime();
        double dWt = (uc + 1.0) * dM;
        double wWt = uc * wM;
        double w = duration * wM / (dWt + wWt);
        calcBottleneckPartials(partials, newPartials);
        for (int i =0; i < uc; i++) {
            System.arraycopy( newPartials, 0, partials, 0, newPartials.length );
            substitutionModel.getTransitionProbabilities(w, branchRate, transProbsMatrix);
            likelihoodCore.calculatePartialsPruning(partials, transProbsMatrix, newPartials);
            System.arraycopy( newPartials, 0, partials, 0, newPartials.length );
            calcBottleneckPartials(partials, newPartials);
        }
    }


    private void checkPartialsMatch(double[] ps1, double[] ps2) {
        for (int i = 0; i < ps1.length; i++) {

            if (Math.abs(ps1[i] - ps2[i]) > 1e-10) {
                System.err.println("Partials mismatch");
                System.exit(1);
            }
        }
    }


    private double[] calcSamplePartials(int iTaxon, int patternCount){
        AlignmentPoMo data = (AlignmentPoMo) dataInput.get();
        int vPop=((nStates-4)/6)+1;
        double[] partials = new double[patternCount * nStates];

        int k = 0;
        int[] nState, maxNucs, maxVals;
        for (int iPattern = 0; iPattern < patternCount; iPattern++) {
            nState = data.getPatternPoMo(iTaxon, iPattern);
            maxNucs = new int[]{-1, -1};
            maxVals = new int[]{0, 0};
            for (int iNuc = 0; iNuc < 4; iNuc++) {
            	if (nState[iNuc]>maxVals[0]){
            		maxNucs[1]=maxNucs[0];
            		maxVals[1]=maxVals[0];
            		maxNucs[0]=iNuc;
            		maxVals[0]=nState[iNuc];
            	}else if(nState[iNuc]>maxVals[1]){
            		maxNucs[1]=iNuc;
            		maxVals[1]=nState[iNuc];
            	}
            }
            if(maxVals[1]>0){
            	int numSeconds=0;
            	boolean equalFirst=false;
            	if (maxVals[1]==maxVals[0])equalFirst=true;
            	for (int iNuc = 0; iNuc < 4; iNuc++) {
            		if(nState[iNuc]==maxVals[1]) numSeconds+=1;
            	}
            	if(numSeconds>1){ //if there are multiple nucleotides sharing the second place of most abundant, than pick one at random as second most abundant.
            		int newSecond=Randomizer.nextInt(numSeconds);
            		int secondCount=0;
            		for (int iNuc = 0; iNuc < 4; iNuc++) {
                		if(nState[iNuc]==maxVals[1] && secondCount==newSecond) maxNucs[1]=iNuc;
                		else secondCount+=1;
                	}
            		if(equalFirst){
            			int newFirst=-1;
            			if(numSeconds==2){ newFirst=1-newSecond;
            			}else{
            				newFirst=Randomizer.nextInt(numSeconds-1);
            				if (newFirst>=newSecond) newFirst+=1;
            				
            			}
            			int firstCount=0;
                		for (int iNuc = 0; iNuc < 4; iNuc++) {
                    		if(nState[iNuc]==maxVals[1] && firstCount==newFirst) maxNucs[0]=iNuc;
                    		else firstCount+=1;
                    	}
            		}
            	}
            }
            if(maxNucs[0]==-1){
            	for (int iState = 0; iState < nStates; iState++) {
                    partials[k++] = 1.0;
                }
            }else {
            	//fixed states
            	for (int iState = 0; iState < 4; iState++) {
            		if (iState==maxNucs[0]){
            			partials[k++] = sampProb(1.0- sequencingError,maxVals[0], sequencingError /3.0,maxVals[1]);
            		} else if (iState==maxNucs[1]){
            			partials[k++] = sampProb(sequencingError /3.0,maxVals[0],1.0- sequencingError,maxVals[1]);
            		} else{
            			partials[k++] = sampProb(sequencingError /3.0,maxVals[0]+maxVals[1],1.0- sequencingError,0);
            		}
            	}
            	//groups of polymorphisms
            	for (int iGroup = 0; iGroup < 6; iGroup++) {
            		int[] ind=PoMoGeneral.indeces(iGroup);
            		for (int iState = 0; iState < vPop-1; iState++) {
                		double[] probs=new double[2];
                		double freq;
                		if (maxNucs[0]==ind[0]) {
                			freq=(iState+1.0)/((double)(vPop));
                			//probs[0]=freq*(1.0-sErr) + (1.0-freq)*sErr/3.0;
                		}else if(maxNucs[0]==ind[1]) {
                			freq=(vPop - (iState+1.0))/((double)(vPop));
                			//probs[0]=(vPop - (iState+1.0))/((double)(vPop));
                		}else{
                			freq=0.0;
                			//probs[0]=sErr;
                		}
                		probs[0]=freq*(1.0- sequencingError) + (1.0-freq)* sequencingError /3.0;
                		if (maxNucs[1]==ind[0]) {
                			freq=(iState+1.0)/((double)(vPop));
                			//probs[1]=(iState+1.0)/((double)(vPop));
                		}else if(maxNucs[1]==ind[1]) {
                			freq=(vPop - (iState+1.0))/((double)(vPop));
                			//probs[1]=(vPop - (iState+1.0))/((double)(vPop));
                		}else{
                			freq=0.0;
                			//probs[1]=sErr;
                		}
                		probs[1]=freq*(1.0- sequencingError) + (1.0-freq)* sequencingError /3.0;
                		partials[k++] = sampProb(probs[0],maxVals[0],probs[1],maxVals[1]);
            		}
            	}
            }
        }
    	return partials;
    }
    
    
    //the partial likelihood of sampling - Nicola
    private double sampProb(double p1, int v1, double p2, int v2){
    	double res=1.0,bigP,smallP;
    	int big,small;
    	if (v1>v2) {
    		big=v1;
    		bigP=p1;
    		small=v2;
    		smallP=p2;
    	}else{
    		big=v2;
    		bigP=p2;
    		small=v1;
    		smallP=p1;
    	}
    			
    	for (int i = 1; i <= small; i++) {
    		res=res*(i+big)*smallP*bigP/i;
    	}
    	for (int i = 1; i <= big-small; i++) {
    		res=res*bigP;
    	}
    	return res;
    }


    private double sumLogPOverPatterns(double [] patternLLhoods) {
        double sumLogP = 0.0;
        if (useAscertainedSitePatterns) {
            final double ascertainmentCorrection = dataInput.get().getAscertainmentCorrection(patternLLhoods);
            assert Double.isFinite(ascertainmentCorrection);
            for (int i = 0; i < nPatterns; i++) {
                sumLogP += (patternLLhoods[i] - ascertainmentCorrection) * dataInput.get().getPatternWeight(i);
                assert Double.isFinite(sumLogP);
            }
        } else {
            for (int i = 0; i < nPatterns; i++) {
                sumLogP += patternLLhoods[i] * dataInput.get().getPatternWeight(i);
                assert Double.isFinite(sumLogP);
            }
        }
        if (sumLogP>0.0){
        	System.out.println("log-likelihood>0, probably matrix exponentiation problem, rejecting move\n");
            sumLogP=Double.NEGATIVE_INFINITY;
        }
        return sumLogP;
    }

    // *****************************************************************************
    // ************************ Making matrices ************************************
    // *****************************************************************************


    private void getTransitionProbabilitiesBottleneck() {
        int i, j, k;
        double temp;

        // is the following really necessary?
        // implemented a pool of iexp matrices to support multiple threads
        // without creating a new matrix each call. - AJD
        // a quick timing experiment shows no difference - RRB
        double[] iexp = new double[nStates * nStates];
        // Eigen vectors
        double[] Evec = eigenDecompositionBottleneck.getEigenVectors();
        // inverse Eigen vectors
        double[] Ievc = eigenDecompositionBottleneck.getInverseEigenVectors();
        // Eigen values
        double[] Eval = eigenDecompositionBottleneck.getEigenValues();
        for (i = 0; i < nStates; i++) {
            temp = Math.exp(bottleneck * Eval[i]);
            for (j = 0; j < nStates; j++) {
                iexp[i * nStates + j] = Ievc[i * nStates + j] * temp;
            }
        }

        int u = 0;
        for (i = 0; i < nStates; i++) {
            for (j = 0; j < nStates; j++) {
                temp = 0.0;
                for (k = 0; k < nStates; k++) {
                    temp += Evec[i * nStates + k] * iexp[k * nStates + j];
                }

                probabilitiesBottleneck[u] = Math.abs(temp);
                u++;
            }
        }
        
        //if exponentiation fails, return matrix of 0s to give -infinite log-likelihood
        for (i = 0; i < nStates; i++) {
        	temp = 0.0;
            for (j = 0; j < nStates; j++) {
                temp += probabilitiesBottleneck[i* nStates +j];
            }
            if (temp>1.001 || temp<0.999){
            	//System.out.println("Matrix exponentiation has given non-probability matrix: rejecting move\n");
            	for (i = 0; i < nStates; i++) {
	            	for (j = 0; j < nStates; j++) {
                        probabilitiesBottleneck[i* nStates +j] = 0.0;
	                }
            	}
            	break;
            }	
        }
    } // getTransitionProbabilities



	private void getTransitionProbabilitiesDrift2nucs() {
    	int dim = virtualPopSize + 1;
		double[] iexp = new double[dim * dim];
		// Eigen vectors
		double[] Evec = eigenDecompositionDrift2nucs.getEigenVectors();
		// inverse Eigen vectors
		double[] Ievc = eigenDecompositionDrift2nucs.getInverseEigenVectors();
		// Eigen values
		double[] Eval = eigenDecompositionDrift2nucs.getEigenValues();
		for (int i = 0; i < dim; i++) {
			double temp = Math.exp(bottleneck * Eval[i]);
			for (int j = 0; j < dim; j++) {
				iexp[i * dim + j] = Ievc[i * dim + j] * temp;
			}
		}
		int u = 0;
		for (int i = 0; i < dim; i++) {
			for (int j = 0; j < dim; j++) {
				double temp = 0.0;
				for (int k = 0; k < dim; k++) {
					temp += Evec[i * dim + k] * iexp[k * dim + j];
				}
                probabilitiesDrift2nucs[u] = Math.abs(temp);
				u++;
			}
		}

		//if exponentiation fails, die
		for (int i = 0; i < dim; i++) {
			double temp = 0.0;
			for (int j = 0; j < dim; j++) {
				temp +=probabilitiesDrift2nucs[i*dim +j];
			}
			assert temp < 1.001 && temp > 0.999;
		}
	}



    
    /**
     * get the PoMo rate Matrix for a population bottleneck.
     * First 4 rows are fixed states, ordered A,C,G,T. Then are polymorhic state group, ordered A->C, A->G, A->T, C->G, C->T, G->T. 
     * Within each group states are ordered 1A-9C, 2A-8C, etc... 
     */
    private double[][] getRatesBottleneck( int vPop){
    	int nrOfStates=6*(vPop-1) +4;
    	double[][] rates;
    	rates= new double[nrOfStates][nrOfStates];
    	int i,j,k;//,k2;
    	for(i=0;i<nrOfStates;i++){
    		for(j=0;j<nrOfStates;j++) rates[i][j]=0.0;
    	}
    	
    	//fixation rates
    	k=0;
    	//k2=0;
    	for(i=0;i<4;i++){
    		for(j=0;j<4;j++){
    			if (i<j){
    				rates[4+k*(vPop-1)][j]=coeff(0.0,0.0, vPop, 1, 0);
    				rates[4+k*(vPop-1)+(vPop-2)][i]=coeff(0.0,0.0, vPop, 1, 0);
    				k+=1;
    			}
    		}	
    	}
    	
    	//other allele frequency changes
    	for(i=0;i<6;i++){
    		//int[] ind=indeces(i);
    		for(j=0;j<vPop-1;j++){
    			if (j>0) rates[4+i*(vPop-1)+j][4+i*(vPop-1)+j-1]=coeff(0.0,0.0, vPop, j+1, j);
    			if (j<vPop-2) rates[4+i*(vPop-1)+j][4+i*(vPop-1)+j+1]=coeff(0.0,0.0, vPop, j+1, j+2);
    		}
    	}
    	
    	//diagonal elements
    	for(i=0;i<nrOfStates;i++){
    		double sum=0.0;
    		for(j=0;j<nrOfStates;j++){
    			if (i!=j)sum+=rates[i][j];
    		}
    		rates[i][i]=-sum;
    	}
    	return rates;
    }
    

    //PoMo allele frequency change rate
    private double coeff(double fit1,double fit2, int vPop, int num1, int num2){
    	double co=0.0;
    	if (num2==num1+1){
    		co=num1*(1.0+fit1-fit2)*(vPop-num1)/(vPop*(num1*(1.0+fit1-fit2) + (vPop-num1)));
    	}else if (num2==num1-1){
    		co=num1*(vPop-num1)/(vPop*(num1*(1.0+fit1-fit2) + (vPop-num1)));
    	}else {
    		System.out.println("Problem with PoMo indeces in calculating the rate matrix");
    		System.exit(0);
    	}
    	return co;
    }


	// Part of the drift-only rate matrix
	private double[][] getDrift2nucsRates(int vPop){
		int dim = vPop + 1;
		double[][] rates = new double[dim][dim];
		for(int i = 1; i < vPop; i++){
			double rate =  (double)i * (double)(vPop - i);
			rate /= (double)(vPop * vPop);
			rates[i][i-1] = rates[i][i+1] = rate;
			rates[i][i] = - 2.0 * rate;
		}
		return rates;
	}


} // class TreeLikelihood
